package org.xbib.interlibrary.client.zdb.sru;

import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.util.Deque;
import java.util.List;
import java.util.Locale;
import java.util.function.Consumer;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.action.resolve.ResolveRequest;
import org.xbib.interlibrary.action.resolve.ResolveResponse;
import org.xbib.interlibrary.action.resolve.ResolveResponseBuilder;
import org.xbib.interlibrary.action.resolve.ResolveServiceResponseBuilder;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.Service;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.common.DefaultBibliographicDescription;
import org.xbib.interlibrary.api.util.Lists;
import org.xbib.interlibrary.sru.AbstractSRUResolveService;
import org.xbib.interlibrary.sru.SRUClientHelper;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.MarcRecordListener;
import org.xbib.marc.MarcXchangeConstants;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.standardnumber.ZDB;

public class SRUResolveService extends AbstractSRUResolveService {

    public SRUResolveService(ServiceArguments arguments, SRUClientHelper helper) {
        super(arguments, helper);
    }

    @Override
    protected String supports(String source) {
        return source != null && source.toLowerCase(Locale.ROOT).contains("zdb") ? "ZDB" : null;
    }

    @Override
    protected void resolve(ResolveRequest resolveRequest,
                           ResolveServiceResponseBuilder serviceResponseBuilder,
                           InputStream inputStream,
                           DefaultBibliographicDescription.Builder descriptionBuilder) {
        MarcContentHandler marcContentHandler = new MarcContentHandler();
        marcContentHandler.setMarcRecordListener(new MarcRecordListener() {
            @Override
            public void beginCollection() {

            }
            @Override
            public void record(MarcRecord marcRecord) {
                String source = getName();
                String sourceId = null; // DNB-ID is not ZDB-ID
                for (MarcField marcField : marcRecord.getAll("016")) {
                    Deque<MarcField.Subfield> typeSubfield = marcField.getSubfield("2");
                    if (typeSubfield != null && !typeSubfield.isEmpty()) {
                        String type = typeSubfield.getFirst().getValue();
                        if ("DE-600".equals(type)) {
                            String id = marcField.getSubfield("a").getFirst().getValue();
                            if (id != null) {
                                source = "ZDB";
                                sourceId = id;
                            }
                        }
                    }
                }
                if (sourceId != null) {
                    // the catalog ID should always be correct. But we recalculate ID here - it may be changed to another ID (for unknown reasons).
                    // Never use checksum fixing, never verify
                    ZDB zdb = new ZDB();
                    zdb.set(sourceId).normalize();
                    sourceId = zdb.normalizedValue();
                    String formattedSourceId = zdb.format();
                    descriptionBuilder.setSource(source);
                    descriptionBuilder.setSourceId(sourceId);
                    descriptionBuilder.setFormattedSourceId(formattedSourceId);
                }
                addIdentifier(descriptionBuilder, "ISBN", marcHelper.extractIBN(marcRecord, "020", null, List.of("a")));
                addIdentifier(descriptionBuilder, "ISSN", marcHelper.extractIBN(marcRecord, "022", null, List.of("a")));
                descriptionBuilder.setInTitle(Lists.firstElementAsStringIfAny(marcHelper.extractTitle(marcRecord, List.of("245"))));
                descriptionBuilder.setPublisher(String.join(", ", marcHelper.extractPublishedBy(marcRecord, List.of("260", "264"), "b")));
                if (resolveRequest.isSourceEnabled()) {
                    serviceResponseBuilder.setSource(marcRecord);
                }
                serviceResponseBuilder.setStatus(getName(), Response.Status.OK);
            }

            @Override
            public void endCollection() {
            }
        });
        try {
            Marc.builder()
                    .setInputStream(inputStream)
                    .setCharset(StandardCharsets.UTF_8)
                    .setContentHandler(new MarcContentHandler()
                            .setFormat("MarcXML")
                            .setType(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE)
                            .addNamespace("http://www.loc.gov/MARC21/slim")
                            .setMarcListener(MarcXchangeConstants.BIBLIOGRAPHIC_TYPE, marcContentHandler)
                            .setMarcListener(MarcXchangeConstants.HOLDINGS_TYPE, marcContentHandler))
                    .build()
                    .xmlReader()
                    .parse();
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    @Override
    public int compareTo(Service<ResolveRequest, ResolveResponse, ResolveResponseBuilder> o) {
        return getName().compareTo(o.getName());
    }

    @Override
    protected String sourceIdCQL(String sourceId, String formattedSourceId) {
        return "dnb.zdbid = " + formattedSourceId;
    }

    @Override
    protected Consumer<TinyMap.Builder<String, Object>> parameterCompletion() {
        return params -> {
            params.put("startRecord", 0);
            params.put("maximumRecords", 1);
        };
    }
}
